@@ -38,7 +38,6 @@ signal.signal(signal.SIGINT, signal_handling)
 
 perceptor, normalize_image = load()
 
-
 # Helpers
 
 def exists(val):
@@ -89,6 +88,19 @@ def norm_siren_output(img):
     return ((img + 1) * 0.5).clamp(0, 1)
 
 
+def create_clip_img_transform(image_width):
+    clip_mean = [0.48145466, 0.4578275, 0.40821073]
+    clip_std = [0.26862954, 0.26130258, 0.27577711]
+    transform = T.Compose([
+                    #T.ToPILImage(),
+                    T.Resize(image_width),
+                    T.CenterCrop((image_width, image_width)),
+                    T.ToTensor(),
+                    T.Normalize(mean=clip_mean, std=clip_std)
+            ])
+    return transform
+
+
 class DeepDaze(nn.Module):
     def __init__(
             self,
@@ -131,7 +143,7 @@ class DeepDaze(nn.Module):
 
         self.generate_size_schedule()
 
-    def forward(self, text, return_loss=True, dry_run=False):
+    def forward(self, text_embed, return_loss=True, dry_run=False):
         out = self.model()
         out = norm_siren_output(out)
 
@@ -151,7 +163,6 @@ class DeepDaze(nn.Module):
 
         with autocast(enabled=False):
             image_embed = perceptor.encode_image(image)
-            text_embed = perceptor.encode_text(text)
 
         if not dry_run:
             self.num_batches_processed += self.batch_size
@@ -205,11 +216,26 @@ class DeepDaze(nn.Module):
         return sizes
 
 
+    
+def create_text_path(text=None, img=None, encoding=None):
+    if text is not None:
+        input_name = text.replace(" ", "_")
+    elif img is not None:
+        if isinstance(img, str):
+            input_name = "".join(img.replace(" ", "_").split(".")[:-1])
+        else:
+            input_name = "PIL_img"
+    else:
+        input_name = "your_encoding"
+    return input_name
+
 class Imagine(nn.Module):
     def __init__(
             self,
-            text,
             *,
+            text=None,
+            img=None,
+            clip_encoding=None,
             lr=1e-5,
             batch_size=4,
             gradient_accumulate_every=4,
@@ -240,6 +266,7 @@ class Imagine(nn.Module):
 
         self.epochs = epochs
         self.iterations = iterations
+        self.image_width = image_width
         total_batches = epochs * iterations * batch_size * gradient_accumulate_every
 
         model = DeepDaze(
@@ -260,9 +287,13 @@ class Imagine(nn.Module):
         self.open_folder = open_folder
         self.save_progress = save_progress
         self.text = text
-        self.textpath = text.replace(" ", "_")
+        self.image = img
+        self.textpath = create_text_path(text=text, img=img, encoding=clip_encoding)
         self.filename = self.image_output_path()
-        self.encoded_text = tokenize(text).cuda()
+
+        # create coding to optimize for
+        self.clip_img_transform = create_clip_img_transform(perceptor.input_resolution.item())
+        self.clip_encoding = self.create_clip_encoding(text=text, img=img, encoding=clip_encoding)
 
         self.start_image = None
         self.start_image_train_iters = start_image_train_iters
@@ -281,6 +312,32 @@ class Imagine(nn.Module):
 
             image_tensor = transform(image)[None, ...].cuda()
             self.start_image = image_tensor
+            
+    def create_clip_encoding(self, text=None, img=None, encoding=None):
+        self.text = text
+        self.img = img
+        if encoding is not None:
+            return encoding.cuda()
+        elif text is not None:
+            return self.create_text_encoding(text)
+        elif img is not None:
+            return self.create_img_encoding(img)
+
+    def create_text_encoding(self, text):
+        tokenized_text = tokenize(text).cuda()
+        text_encoding = perceptor.encode_text(tokenized_text).detach()
+        return text_encoding
+    
+    def create_img_encoding(self, img):
+        if isinstance(img, str):
+            img = Image.open(img)
+        normed_img = self.clip_img_transform(img).unsqueeze(0).cuda()
+        img_encoding = perceptor.encode_image(normed_img).detach()
+        return img_encoding
+    
+    def set_clip_encoding(self, text=None, img=None, encoding=None):
+        encoding = self.create_clip_encoding(text=text, img=img, encoding=encoding).cuda()
+        self.clip_encoding = encoding
 
     def image_output_path(self, sequence_number=None):
         """
@@ -298,26 +355,12 @@ class Imagine(nn.Module):
             output_path = f"{current_time}_{output_path}"
         return Path(f"{output_path}.png")
 
-    def generate_and_save_image(self, sequence_number=None):
-        """
-        :param sequence_number:
-        :param custom_filename: A custom filename to use when saving - e.g. "testing.png"
-        """
-        with torch.no_grad():
-            img = normalize_image(self.model(self.encoded_text, return_loss=False).cpu())
-            img.clamp_(0., 1.)
-            self.filename = self.image_output_path(sequence_number=sequence_number)
-            save_image(img, self.filename)
-            save_image(img, f"{self.textpath}.png")
-
-            tqdm.write(f'image updated at "./{str(self.filename)}"')
-
-    def train_step(self, epoch, iteration) -> int:
+    def train_step(self, epoch, iteration):
         total_loss = 0
 
         for _ in range(self.gradient_accumulate_every):
             with autocast():
-                loss = self.model(self.encoded_text)
+                loss = self.model(self.clip_encoding)
             loss = loss / self.gradient_accumulate_every
             total_loss += loss
             self.scaler.scale(loss).backward()
@@ -327,12 +370,23 @@ class Imagine(nn.Module):
         self.optimizer.zero_grad()
 
         if (iteration % self.save_every == 0) and self.save_progress:
-            current_total_iterations = epoch * self.iterations + iteration
-            sequence_number = current_total_iterations // self.save_every
-            self.generate_and_save_image(sequence_number=sequence_number)
+            self.save_image(epoch, iteration)
 
         return total_loss
 
+    @torch.no_grad()
+    def save_image(self, epoch, iteration):
+        current_total_iterations = epoch * self.iterations + iteration
+        sequence_number = current_total_iterations // self.save_every
+
+        img = normalize_image(self.model(self.clip_encoding, return_loss=False).cpu())
+        img.clamp_(0., 1.)
+        self.filename = self.image_output_path(sequence_number=sequence_number)
+        save_image(img, self.filename)
+        save_image(img, f"{self.textpath}.png")
+
+        tqdm.write(f'image updated at "./{str(self.filename)}"')
+
     def forward(self):
         if exists(self.start_image):
             tqdm.write('Preparing with initial image...')
@@ -353,9 +407,9 @@ class Imagine(nn.Module):
             del self.start_image
             del optim
 
-        tqdm.write(f'Imagining "{self.text}" from the depths of my weights...')
+        tqdm.write(f'Imagining "{self.textpath}" from the depths of my weights...')
 
-        self.model(self.encoded_text, dry_run = True) # do one warmup step due to potential issue with CLIP and CUDA
+        self.model(self.clip_encoding, dry_run = True) # do one warmup step due to potential issue with CLIP and CUDA
 
         if self.open_folder:
             open_folder('./')
@@ -367,10 +421,8 @@ class Imagine(nn.Module):
                 loss = self.train_step(epoch, i)
                 pbar.set_description(f'loss: {loss.item():.2f}')
 
-                if self.model.num_batches_processed > self.model.total_batches:
-                    print('number of batches processed exceeds calculated total batches')
-                    return
-
                 if terminate:
                     print('interrupted by keyboard, gracefully exiting')
                     return
+
+        self.save_image(self.epochs, self.iterations) # one final save at end
